Package: unitizer
Title: Interactive R Unit Tests
Description: Simplifies regression tests by comparing objects produced by test
    code with earlier versions of those same objects.  If objects are unchanged
    the tests pass, otherwise execution stops with error details.  If in
    interactive mode, tests can be reviewed through the provided interactive
    environment.
Version: 1.4.23
Authors@R: c(
    person(
    "Brodie", "Gaslam", email="brodie.gaslam@yahoo.com",
    role=c("aut", "cre")),
    person(
    "Michael", "https://github.com/MichaelChirico", 
    email="michaelchirico4@gmail.com",
    role=c("ctb")),
    person(family="R Core Team",
    email="R-core@r-project.org", role="cph",
    comment="Traceback function sources."
    )    )
Depends: methods
Imports: stats, utils, crayon (>= 1.3.2), diffobj (>= 0.1.5.9000)
VignetteBuilder: knitr
Suggests: knitr, rmarkdown
License: GPL-2 | GPL-3
URL: https://github.com/brodieG/unitizer
BugReports: https://github.com/brodieG/unitizer/issues
Collate: 'asciiml.R' 'capture.R' 'is.R' 'global.R' 'change.R'
        'class_unions.R' 'list.R' 'conditions.R' 'item.R' 'deparse.R'
        'text.R' 'item.sub.R' 'section.R' 'test.R' 'unitizer.R'
        'exec.R' 'prompt.R' 'browse.struct.R' 'browse.R' 'demo.R'
        'diff.R' 'faux_prompt.R' 'get.R' 'heal.R' 'load.R' 'ls.R'
        'misc.R' 'search.R' 'options.R' 'onload.R' 'parse.R' 'rename.R'
        'repairenvs.R' 'result.R' 'shims.R' 'size.R' 'state.R'
        'state.compare.R' 'traceback.R' 'translate.R' 'unitize.R'
        'unitize.core.R' 'unitizer-package.R' 'unitizer.add.R'
        'upgrade.R'
RoxygenNote: 7.3.3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2025-10-21 21:24:54 UTC; brodie
Author: Brodie Gaslam [aut, cre],
  Michael https://github.com/MichaelChirico [ctb],
  R Core Team [cph] (Traceback function sources.)
Maintainer: Brodie Gaslam <brodie.gaslam@yahoo.com>
Repository: CRAN
Date/Publication: 2025-10-22 07:30:14 UTC
Built: R 4.6.0; ; 2026-04-25 23:01:35 UTC; unix
