Package: utzflm
Title: Quickly Compute Simple Linear Regressions
Description: This is a demo package used to illustrate usage of the `unitizer`
    package.
Version: 0.1.0
Author: Brodie Gaslam
Maintainer: Brodie Gaslam <brodieDOTgaslam@MARISSAMAYERSCOMPANY.com>
Depends: R (>= 3.0.2)
Imports: stats
Suggests: unitizer
License: GPL-2
LazyData: true
RoxygenNote: 5.0.1
NeedsCompilation: no
Packaged: 2016-10-27 15:16:21 UTC; milberg
Built: R 3.3.1; ; 2016-10-27 15:16:30 UTC; unix
