# Generated by roxygen2: do not edit by hand

S3method(all.equal,condition)
S3method(all.equal,conditionList)
S3method(all.equal,unitizerDummy)
S3method(all.equal,unitizerStateRaw)
S3method(all.equal,unitizer_glob_state_test)
S3method(as.character,H1)
S3method(as.character,H2)
S3method(as.character,H3)
S3method(as.character,bullet)
S3method(as.character,header)
S3method(as.list,unitizerList)
S3method(get_unitizer,character)
S3method(get_unitizer,default)
S3method(get_unitizer,unitizer_error_store)
S3method(get_unitizer,unitizer_result)
S3method(get_unitizer,unitizer_results)
S3method(infer_unitizer_location,character)
S3method(infer_unitizer_location,default)
S3method(print,bullet)
S3method(print,captured_output)
S3method(print,header)
S3method(print,unitizer_ls)
S3method(print,unitizer_result)
S3method(print,unitizer_results)
S3method(set_unitizer,character)
S3method(set_unitizer,default)
export("[Press ENTER to Continue]")
export(all_eq)
export(append)
export(conditionList)
export(copy_fastlm_to_tmpdir)
export(desc)
export(editCalls)
export(filename_to_storeid)
export(get_unitizer)
export(in_pkg)
export(infer_unitizer_location)
export(mock_item)
export(repair_environments)
export(review)
export(set_unitizer)
export(show_file)
export(state)
export(testFuns)
export(testthat_translate_dir)
export(testthat_translate_file)
export(testthat_translate_name)
export(unitize)
export(unitize_dir)
export(unitizer_check_demo_state)
export(unitizer_cleanup_demo)
export(unitizer_sect)
export(update_fastlm)
exportClasses(conditionList)
exportMethods("$")
exportMethods("[[")
exportMethods(all.equal)
exportMethods(editCalls)
exportMethods(healEnvs)
exportMethods(show)
import(diffobj)
import(methods)
import(stats)
importFrom(utils,capture.output)
importFrom(utils,file_test)
importFrom(utils,getParseData)
importFrom(utils,installed.packages)
importFrom(utils,loadhistory)
importFrom(utils,modifyList)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,remove.packages)
importFrom(utils,savehistory)
importFrom(utils,sessionInfo)
