<!-- README.md is generated from README.Rmd. Please edit that file 

# render('README.Rmd', output_format=html_vignette(css='vignettes/styles.css'))
rmarkdown::render('README.Rmd', output_format=rmarkdown::md_document())

-->

# unitizeR - Interactive R Unit Tests

[![R build status](https://github.com/brodieG/unitizer/workflows/R-CMD-check/badge.svg)](https://github.com/brodieG/unitizer/actions)
[![](https://codecov.io/github/brodieG/unitizer/coverage.svg?branch=rc)](https://app.codecov.io/gh/brodieG/unitizer?branch=rc)
[![](http://www.r-pkg.org/badges/version/unitizer)](https://cran.r-project.org/package=unitizer)
[![Dependencies direct/recursive](https://tinyverse.netlify.app/badge/unitizer)](https://tinyverse.netlify.app/)
[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)

```{r child='vignettes/rmdhunks/intro.Rmd'}
```

```{r child='vignettes/rmdhunks/usage.Rmd'}
```

## Documentation

* `help(package="unitizer")`, in particular `?unitize`
* `demo(package="unitizer")`
* [`browseVignettes("unitizer")`](https://cran.r-project.org/package=unitizer/vignettes/u0_unitizer_index.html)
  for a list of vignettes, or skip straight to the [Introduction
  vignette](https://cran.r-project.org/package=unitizer/vignettes/u1_intro.html)

## Related Packages

* [`aammrtf`](https://github.com/brodieG/aammrtf), a minimal version of
  `unitizer`, used by `unitizer` for its own tests.
* [`testthat`](https://cran.r-project.org/package=testthat).
* [`tinytest`](https://cran.r-project.org/package=tinytest), which is extended
  by [`ttdo`](https://cran.r-project.org/package=ttdo) for
  [`diffobj`](https://cran.r-project.org/package=diffobj) diffs.
* [`RUnit`](https://cran.r-project.org/package=RUnit).

## Acknowledgments

Thank you to:

* R Core for developing and maintaining such a wonderfully language.
* CRAN maintainers, for patiently shepherding packages onto CRAN and maintaining
  the repository, and Uwe Ligges in particular for maintaining
  [Winbuilder](https://win-builder.r-project.org/).
* [Gábor Csárdi](https://github.com/gaborcsardi) for
  [crayon](https://cran.r-project.org/package=crayon) through which we can
  add a new dimension to the R experience.
* [Jim Hester](https://github.com/jimhester) because
  [covr](https://cran.r-project.org/package=covr) rocks.
* [Hadley Wickham](https://github.com/hadley) for
  [testthat](https://cran.r-project.org/package=testthat) from which we borrow
  many concepts, and for his many other packages.
* [Dirk Eddelbuettel](https://github.com/eddelbuettel) and [Carl
  Boettiger](https://github.com/cboettig) for the
  [rocker](https://github.com/rocker-org/rocker) project, and [Gábor
  Csárdi](https://github.com/gaborcsardi) and the R-consortium for
  [Rhub](https://github.com/r-hub/rhub), without which testing bugs on R-devel
  and other platforms would be a nightmare.
* [Yihui Xie](https://github.com/yihui) for
  [knitr](https://cran.r-project.org/package=knitr) and  [J.J.
  Allaire](https://github.com/jjallaire) et al. for
  [rmarkdown](https://cran.r-project.org/package=rmarkdown), and by extension
  John MacFarlane for [pandoc](https://pandoc.org/).
* @kohler for [gifsicle](https://github.com/kohler/gifsicle) and the [ffmpeg
  team](http://ffmpeg.org/about.html) for ffmpeg.
* All open source developers out there that make their work freely available
  for others to use.
* [Github](https://github.com/), [Codecov](https://about.codecov.io/),
  [Vagrant](https://www.vagrantup.com/), [Docker](https://www.docker.com/),
  [Ubuntu](https://ubuntu.com/), [Brew](https://brew.sh/) for providing
  infrastructure that greatly simplifies open source development.
* [Free Software Foundation](https://www.fsf.org/) for developing the GPL license and
  promotion of the free software movement.

## About the Author

Brodie Gaslam is a hobbyist programmer based in the US East Coast.
