# Update this file with any common steps you wish all your test files to use,
# and add e.g. `source('aammrtf/init.R')` to each of them.

options(
  warn=1,                      # don't accumulate warnings
  useFancyQuotes=FALSE,        # all.equals uses fancy quotes
  encoding="UTF-8",            # Assume UTF-8 input in non-UTF-8 locales
  warnPartialMatchArgs=TRUE,
  warnPartialMatchAttr=TRUE,
  warnPartialMatchDollar=TRUE
)
