% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/size.R
\name{sizeRDS}
\alias{sizeRDS}
\title{Measure object size as an RDS}
\usage{
sizeRDS(object)
}
\description{
Measure object size as an RDS
}
\keyword{internal}
