% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/change.R
\name{length,unitizerChanges-method}
\alias{length,unitizerChanges-method}
\title{Return Sum of Total Changes}
\usage{
\S4method{length}{unitizerChanges}(x)
}
\description{
Return Sum of Total Changes
}
\keyword{internal}
