% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shims.R
\name{getFun}
\alias{getFun}
\title{Utility Function}
\usage{
getFun(name)
}
\description{
Utility Function
}
\keyword{internal}
