% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unitize.core.R
\name{validate_pre_post}
\alias{validate_pre_post}
\title{Helper function for validations}
\usage{
validate_pre_post(what, test.dir)
}
\description{
Helper function for validations
}
\keyword{internal}
